#include "saw.h"
#include "sawdoubler.h"
#include "lattice.h"
#include "symmetry.h"

/* These are the lattice-dependent functions for the 3D cubic lattice */

void go(int i, int K, int N, int ctype, int *visited, int *lattice, int *r,
        cint *nwalks, struct gridpoint *total, struct treenode *Nodes, nint *nodes,
        int option, int *Term, int nt, int rT, int *distT, int *occursT,
        int *EnoughMem, int a, int b, int c) {

    /* This recursive function creates all SAWs of length K which 
       start with a self-avoiding partial walk
            (x[0],y[0],z[0]),..., (x[i],y[i],z[i]),
       where 0 <= i <= K. The current walk is stored in the tree.
       If option = 0, it is always stored in the tree.
       If option = 1, it is only stored if it goes through all the terminal sites
                      Term[i], 0 <= i < nt. The highest terminal site is
                      T = Term[0], encoded by the canonical number rT.
                      The boolean variable occursT registers whether the walk
                      goes through T[0]. The walk is stored without
                      the terminal sites.
       If option = 2, as in option 1, but the walk is only counted, not stored

       Each SAW is added to the tree Nodes, with nodes the current number
       of nodes in the tree. nwalks is the total number of SAWs encountered.
       The number of SAWs stored may be less than nwalks. 
       total stores the total sum of the endpoint coordinates (x,y,z)
       and the squared endpoint distance (rr=r^2).

       visited[r] registers whether point r has been visited already;
       lattice[r] gives its site number;
       distT[r] is its distance to site T, which is used to cut off walks
                that cannot reach T. 

       a, b, c are site numbers defining ranges used to filter walks
       to be stored in the tree.
  
    */

       int i1, j, L, canon_N, delta, n, ri, occursT_tmp;
       int Walk[MAXN] ;

       ri = r[i] ;
       occursT_tmp = *occursT ;
       L = 2*N + 1 ;

       /* If the new site equals the highest from Term, register its occurrence */
       if (option>=1){
           if (ri==rT)
               *occursT = TRUE; 

           /* If T does not occur and cannot be reached anymore, then backtrack */
           if (!(*occursT) && distT[ri] > K-i)
               return ;
       }

       /* If i=K, we have a SAW of length K */
       if (i==K){
           /* Register end point coordinates of walk */
           int L2 = L*L ;
           struct gridpoint endpoint ;

           endpoint.x = r[K]/L2 - N ;
           endpoint.y = (r[K]/L)%L - N ;
           endpoint.z = r[K]%L - N ;
           endpoint.rr = endpoint.x * endpoint.x + endpoint.y * endpoint.y + endpoint.z * endpoint.z ;

           if ( (option>=1 && *occursT) || option==0 ){

               /* Register walk starting from 1 */
               for (j=1; j<=K; j++)
                   Walk[j-1] = lattice[r[j]] ;

               n = K ;
               if (option==0)
                   InsertionSort(Walk, K) ;
               else
                   n = FilterSort_Walk(Walk, K, Term, nt, a, b, c) ;
               if (n >= 0){
                   (*nwalks)++ ;
                   total->x += endpoint.x ;
                   total->y += endpoint.y ;
                   total->z += endpoint.z ;
                   total->rr += endpoint.rr ;
               }
               if (n > 0 && (option==1  || option==0) && *EnoughMem){
                   if (!AddWalkToTree(Walk, n, ctype, endpoint, Nodes, nodes))
                       *EnoughMem = FALSE ;
               }    
               *occursT = occursT_tmp ;
               return ;
           } else if (option==-1) {
               /* Just count the walk, do not add it to the tree */
               (*nwalks)++ ;
               total->x += endpoint.x ;
               total->y += endpoint.y ;
               total->z += endpoint.z ;
               total->rr += endpoint.rr ;
               return ;
           }
       }

       /*** Visit all 6 neighbours ***/

       visited[ri] = TRUE;
       i1 = i+1 ;

      /* canonical numbers differ by delta = -1, +1, -L, +L, -L*L, +L*L */
       canon_N = Canon(N) ;
       delta = -1;
       for (j=0; j<6; j++){
           if (0 <= ri+delta && ri+delta < canon_N){
               r[i1] = ri + delta ;
               if (!visited[r[i1]]){
                   go(i1,K,N,ctype,visited,lattice,r,nwalks,total,Nodes,nodes,option,
                      Term,nt,rT,distT,occursT,EnoughMem,a,b,c);
               }
           }
           delta = -delta ;
           if (delta < 0)
               delta *= L ;
       }

       visited[ri] = FALSE ;
       *occursT = occursT_tmp ;
        
       return ;
    
} /* end go */


void InitDist(int N, int T, int *rl, int *distT){

    /* This function initialises the distance in the 3D cubic lattice from
       all the lattice points to the termination site T.

       Input: N = maximum walk length for the lattice, 
              T = termination site,
              rl[s] gives the canonical number of site s.

       Output: distT[r] = distance to T of lattice point with canonical number r.
    */

       int i, j, k, L, L2, xT, yT, zT, rT, r ;
       int absi, absij ;

       L = 2*N + 1 ;
       L2 = L*L ;

       /* Translate site T to 3D coordinates */
       rT = rl[T] ;
       xT = rT/L2 ; 
       rT -= xT*L2 ;
       yT = rT/L ;
       zT = rT - yT*L ;

       xT -= N ; yT -= N ; zT -= N ;

       /* Compute distance to T */
       for (i=-N; i<=N; i++){
           absi = abs(i) ;
           for (j=-N+absi; j<=N-absi; j++){
               absij = absi + abs(j) ;
               for (k=-N+absij; k<=N-absij; k++){
                   r = ((i+N)*L + j+N )*L + k+N ;
                   distT[r] = abs(xT-i) + abs(yT-j) + abs(zT-k);
               }
           }
       }

       return ;

} /* end InitDist */


int numbering(int N, int *lattice, int *rl) {

  /* This function creates a site numbering for the 3D cubic lattice, 
     where lattice points related by symmetry have the same site number div SYMM.
     The sites are ordered by increasing Euclidean distance from the origin.
     Only the reachable points (i,j,k) with |i|+|j|+|k| <= N obtain
     a non-dummy value. All others get a dummy value.

     For a lattice point (i,j,k) with canonical number
         r = (i+N)*L*L + (j+N)*L + (k+N)
     where L = 2N+1, the site number is given by lattice[r].
     Inversely, for a given site s, the canonical number is given by rl[s].
     Hence, lattice[rl[s]] = s for all assigned site numbers s.
     Also, rl[lattice[r]] = r for all reachable points r.

     The function returns the maximum site number assigned.
  */

     int num, maxnum ;
     int i, j, k, r, r2, L, absk, abskj ;
     
     L = 2*N + 1 ;
    
     /* Initialise lattice to dummy values */
     for (i=0; i<Canon(N); i++)
         lattice[i] = DUMMY ;
                 
     /* Assign site numbers for points (i,j,k) with 0 <= i <= j <= k <= N
        that can be reached by a walk of length N */
     num = 0;
     maxnum = 0;
     for (r2=0; r2<=N*N; r2++){
         /* Assign sites for points with i^2+j^2+k^2 = r2 */ 
         for (k=0; k*k <= r2; k++){
             for (j=0; j<=k && j*j+k*k <= r2 ; j++){
                 for (i=0; i<=j && i*i+j*j+k*k <= r2; i++){
                     if (i+j+k<=N && i*i+j*j+k*k == r2) {
                         lattice[((i+N)*L + j+N)*L + k+N] = num ;
                         maxnum = num ;
                         num += SYMM ;
                         break ;
                     }
                 }
             }
         }
     }

     /* Initialise rl to dummy values */
     for (i=0; i<maxnum+SYMM; i++)
         rl[i] = DUMMY ;

     /* Assign numbers for points (i,j,k) with -N <= i,j,k <= N that
        can be reached by a walk of length N. Also assign inverse map rl. */
     for (k=-N; k<=N; k++){
         absk = abs(k) ;
         for (j=-N+absk; j<=N-absk; j++){
             abskj = absk + abs(j) ;
             for (i=-N+abskj; i<=N-abskj; i++){
                 r = ((i+N)*L + j+N)*L + k+N ;
                 num = val(i, j, k, 0, N, lattice) ;
                 lattice[r] = num ;
                 rl[num] = r ;
                 if (num > maxnum)
                     maxnum = num ;
             } 
         }
     }
             
     return maxnum ;

} /* end numbering */
