#include "saw.h"
#include "lattice.h"
#include "symmetry.h"
#include "iteration.h"

/*  SAWdoubler is a software package implementing the length-doubling method
    described in:
        "Exact enumeration of self-avoiding walks"
        by Raoul D. Schram, Gerard T. Barkema, and Rob H. Bisseling
        Journal of Statistical Mechanics: Theory and Experiment (2011) p06019.

    Detailed explanations of the implementation can be found in: 
        "SAWdoubler: a program for counting self-avoiding walks"
        by Raoul D. Schram, Gerard T. Barkema, and Rob H. Bisseling
        Computer Physics Communications, vol. 184 (2013), pp. 891-898.

    The main program of SAWdoubler, called saw, creates all self-avoiding walks
    of length N in 3D, and translates the lattice coordinates of visited sites
    to site numbers.

    Each walk is added to a tree (Nodes), in order of increasing site number.
    The program then recursively computes all subsets S of sites contained 
    in a walk, counts the number of walks Z(S) that go through each S,
    and sums the square of this count, multiplied by the sign (-1)^|S|. 
    Adding this correction term to (Z_N)^2 gives the total count 
    Z_2N of the number of walks of length 2N.

    The program also computes the total squared endpoint distance P_2N. 

    To count all walks of length 2N, for N>=1, the program is used as 
        saw N                           

    The program can also be used to count the number of walks of length M+N,
    using all walks of length N and those of length M, with M, N >= 1,
    by the command
        saw M N
*/

int main(int argc, char **argv) {

    int *visited ;       /* visit status of the lattice */
    int *lattice ;       /* site numbering of the lattice */ 
    int *distT ;         /* distance to site T */
    int *rl ;            /* rl[s] is the canonical number of site number s */
    int r[MAXN+1] ;      /* r[i] is the canonical number of the i-th point of the walk */

    struct treenode *Nodes ; /* tree data structure */
    struct vertex *V ;   /* bookkeeping array for undo operations */
    struct bin *Bins ;   /* data structure for active vertices */

    int Term[MAXN+1] ;   /* set of termination sites */

    int N, M ;           /* length of walk */
    int T ;              /* termination site T > 0, T=Term[0] */
    int nt ;             /* number of termination sites */
    int rT ;             /* location of site T in canonical numbering */
    int T0, T1 ;         /* range T0 <= T <= T1 */
    int occursT ;        /* TRUE if site T occurs in the walk */
    int maxnum ;         /* maximum reachable site number of lattice */
    ccint Z_2N ;         /* number of walks of length M+N */
    ccint Z_N, Z_M ;     /* number of walks of length N, M */
    cint nwalks ;        /* number of walks created */
    ccint P_2N ;         /* sum of squared endpoint distance of walks of length M+N */
    ccint P_N, P_M ;     /* sum of squared endpoint distance of walks created of length N, M */
    struct gridpoint total ; /* total sum of the endpoint coordinates
                                and the squared endpoint distance */
    nint nodes ;         /* number of nodes in the tree */
    nint maxnodes ;      /* maximum number of nodes in all trees */
    nint max_nincludes ; /* maximum number of includes in all trees */
    nint niters ;        /* number of iterations of main program */
    int EnoughMem ;      /* enough memory to store the tree */
    int i ;

#ifdef LONGRUN
    FILE *fplong ;
#endif

    if (argc < 2){ 
        printf("Usage: %s N M\n", argv[0]) ;
        printf("\t-N,M: walk lengths\n");
        exit(0) ;
    }

    /* Read N from command line */
    sscanf(argv[1],"%d",&N);
    printf("Length of a walk = %d\n", N) ;

    if (argc >= 3){
        sscanf(argv[2],"%d",&M);
        printf("Length of second walk = %d\n", M) ;
    } else
        M = 0 ;   /* this means M=N, so we use only one walk length */

    /* Ensure that M <= N, since max(M,N) determines the lattice numbering */
    if (N < M){
        int tmp ;
        tmp = N;
        N = M ; 
        M = tmp ;
    }

    Nodes  = (struct treenode *) malloc(MAXNODES * sizeof(struct treenode));
    V = (struct vertex *) malloc(MAXNODES * sizeof(struct vertex));
    Bins = (struct bin *) malloc(MAXSITES * sizeof(struct bin));
    rl = (int *) malloc((MAXSITES+SYMM+1) * sizeof(int));
    visited = (int *)  malloc(Canon(N)*sizeof(int));
    lattice = (int *)  malloc(Canon(N)*sizeof(int));
    distT   = (int *)  malloc(Canon(N)*sizeof(int));

    if (Nodes == NULL || V == NULL || Bins == NULL || rl == NULL ||
        visited == NULL || lattice == NULL || distT == NULL ) {
        fprintf(stderr, "main(): Not enough memory!\n Please reduce the maximum number of nodes in saw.h\n");
        exit(0);
    }

    /* Initialisations */
    nodes = 0;
    maxnodes = 0;
    max_nincludes = 0 ;
    niters = 0 ;
    EnoughMem = TRUE ;

    /* Number the reachable sites and initialise the others to a dummy value */
    maxnum = numbering(N, lattice, rl) ;
    printf("Maximum site number = %d\n", maxnum) ;
    
    /* Create all walks of length N (Type 1) and count them */
    r[0] = Origin(N) ; /* all walks start at the origin */
    occursT = FALSE ;
    nwalks = 0 ;
    InitPoint(&total) ;

    for (i=0; i<Canon(N) ; i++)
        visited[i] = FALSE ;
    go(0, N, N, 1, visited, lattice, r, &nwalks, &total, Nodes, &nodes,
       -1, Term, 0, DUMMY, distT, &occursT, &EnoughMem,-1,-1,-1) ;
    printf("Total number of walks Z[%d] = ", N) ;
    Z_N = nwalks ;
    print_verylong(stdout, Z_N) ; printf("\n") ;

    printf("Total squared endpoint distance P[%d] = ", N) ;
    P_N = total.rr;
    print_verylong(stdout, P_N) ; printf("\n") ;

    
    if (M > 0){
        /* Create all walks of length M (Type 2) and count them */
        r[0] = Origin(N) ;
        occursT = FALSE ;
        nwalks = 0 ;
        InitPoint(&total) ;
        for (i=0; i<Canon(N) ; i++)
            visited[i] = FALSE ;
        go(0, M, N, 2, visited, lattice, r, &nwalks, &total, Nodes, &nodes,
           -1, Term, 0, DUMMY, distT, &occursT, &EnoughMem,-1,-1,-1) ;
        printf("Total number of walks Z[%d] = ", M) ;
        Z_M = nwalks ;
        print_verylong(stdout, Z_M) ; printf("\n") ;

        printf("Total squared endpoint distance P[%d] = ", M) ;
        P_M = total.rr ;
        print_verylong(stdout, P_M) ; printf("\n") ;
    } else {
        Z_M = Z_N ;
        P_M = P_N ;
    }

    Z_2N = Z_M * Z_N ;
    printf("Z_2N = ") ;
    print_verylong(stdout, Z_2N) ; printf("\n") ;

    P_2N = Z_M * P_N + Z_N * P_M ;
    printf("P_2N = ") ;
    print_verylong(stdout, P_2N) ; printf("\n") ;

    for (T0 = 0; T0 <= maxnum; T0 += SYMM) {
        /* Find highest occurring site number T with T0 <= T < T0+SYMM */
        if (T0+SYMM-1 <= maxnum)
            T1 = T0+SYMM-1 ;
        else
            T1 = maxnum ;

        rT = DUMMY ;
        for (i = T1; i>= T0; i--) {
            if (rl[i] != DUMMY) {
                T = i ;
                rT = rl[i] ;
                break ;
            }
        }
        if (rT == DUMMY){
            printf("skipped T0 = %d\n", T0);
            continue ;
        }
 
        /*** Handle all subwalks with T as highest occurring site ***/
  
        /* Initialise distances to T */
        InitDist(N, T, rl, distT) ;
        
        /* Perform one iteration */
        nt = 1 ;
        Term[0] = T;
        Iteration(M, N, visited, lattice, r, rl, Nodes, Bins, V,
                  Term, nt, rT, distT, maxnum, &maxnodes, &max_nincludes, 
                  &niters, &Z_2N, &P_2N, -1, -1) ;

        printf("T0 = %d    Z_2N = ", T0) ;
        print_verylong(stdout, Z_2N) ; printf("\n") ;
        printf("           P_2N = ") ;
        print_verylong(stdout, P_2N) ; printf("\n") ;

#ifdef LONGRUN
        /* Make sure that an intermediate result is really printed */
        fplong=fopen("logfile","a");
        fprintf(fplong, "T0 = %d    Z_2N = ", T0) ;
        print_verylong(fplong, Z_2N);
        fprintf(fplong, "\n");
        fprintf(fplong, "           P_2N = ") ;
        print_verylong(fplong, P_2N);
        fprintf(fplong, "\n");
        fclose(fplong);
#endif

    }

    /*** Output the results ***/

    printf("Maximum number of nodes in a tree = ") ;
    print_verylong(stdout, (ccint)maxnodes) ; printf("\n") ;
    
    printf("Maximum number of includes for a tree = ") ;
    print_verylong(stdout, (ccint)max_nincludes) ; printf("\n") ;

    printf("Number of iterations = ") ;
    print_verylong(stdout, (ccint)niters) ; printf("\n") ;
    
    if (M==0)
        printf("Total number of walks Z[%d] = ", 2*N) ;
    else
        printf("Total number of walks Z[%d] = ", M+N) ;
    print_verylong(stdout, Z_2N) ; printf("\n") ;
    if (M==0)
        printf("Total squared endpoint distance P[%d] = ", 2*N) ;
    else
        printf("Total squared endpoint distance P[%d] = ", M+N) ;
    print_verylong(stdout, P_2N) ; printf("\n") ;

    if (distT != NULL)
        free(distT) ;
    if (lattice != NULL)
        free(lattice) ;
    if (visited != NULL)
        free(visited) ;
    if (rl != NULL)
        free(rl) ;
    if (Bins != NULL)
        free(Bins) ;
    if (V != NULL)
        free(V) ;
    if (Nodes != NULL)
        free(Nodes) ;

    exit(0) ;

} /* end main */
