#include "saw.h"
#include "lattice.h"
#include "symmetry.h"
#include "sawdoubler.h"

void InitTree(struct treenode *Nodes, nint *nodes) {

    /* This function initialises a tree representing self-avoiding walks.
       On output, the tree contains the root, node 0, which represents 
       the origin. The number of nodes is set to 1. Only the fields 
       representing the tree structure and the count field are initialised.
    */

       Nodes[0].Site = ROOT ;
       InitNodeCounts (&(Nodes[0])) ;
       Nodes[0].Child = DUMMY ;
       Nodes[0].Sibling = DUMMY ;
       Nodes[0].Parent = DUMMY ;
       *nodes = 1 ;

} /* end InitTree */


int AddWalkToTree(int *Walk, int n, int ctype, struct gridpoint endpoint,
                  struct treenode *Nodes, nint *nodes) {

    /* This function adds a walk of length n>=1 to the tree,
       with a given walk count. The walk starts at site 0
       and visits sites Walk[0], Walk[1], ..., Walk[n-1],
       though not necessarily in that order. The sites of the walk
       in the array Walk are assumed ordered by increasing site number,
       Walk[0] < Walk[1] < ... < Walk[n-1]. The walk count gives the
       number of different walks which visit the same sites and is  
       stored with site Walk[n-1]. Counts can be stored for 2 types of walks:
           Type 1 stored in Nodes[i].Count
           Type 2 stored in Nodes[i].Count2
       Type 1 can e.g. have a longer length than Type 2.
       The variable ctype=1,2 gives the type to be added to the tree.

       The tree is a data structure storing information on all the walks
       in an easily accessible format, which avoids repetition of
       lower-numbered parts of the walk. The tree consists of nodes i.
       Each node i has a corresponding site in the lattice, given by the 
       number Nodes[i].Site.
       
       Each path from the root to a leaf i in the tree represents one
       or more walks passing through the same sites. The walks stored 
       in the tree need not have the same length. Nodes[i].Count represents 
       the number of walks from the root to node i, ending in node i,
       i.e., with Nodes[i].Site as the highest site number.

       The tree structure is given by 3 pointers for each node i:

           Nodes[i].Child is the first child
           Nodes[i].Sibling is the next sibling
           Nodes[i].Parent is the parent

       Node 0 is the root, which corresponds to site ROOT.
       If there is no child, sibling, or parent, the corresponding
       pointer is set to DUMMY. The siblings are ordered by increasing
       site number.

       nodes is the number of nodes of the tree after addition of the walk. 
    */

       nint i, previ=DUMMY, parent, t ;
       int j ;

       if (n < 1)
           return TRUE ;

       parent = 0 ; /* root node */
       i = Nodes[0].Child ;
       j = 0 ; 
       while ( j < n ) {
           /*** Add site Walk[j] ***/

           if (i != DUMMY){
               /* Check siblings to find Walk[j] */
               previ = i;
               while ((Nodes[i].Site < Walk[j]) && (Nodes[i].Sibling != DUMMY) ) {
                   previ = i;
                   i = Nodes[i].Sibling ;
               }
           }
           
           if (i != DUMMY && Nodes[i].Site == Walk[j]) {
               /* Found: Walk[0], ... , Walk[j] are already in tree */
               if (j == n-1){
                   if (ctype == 1){
                       Nodes[i].Count++ ;
                   } else if (ctype == 2){
                       Nodes[i].Count2++ ;
                   }
                   AddCountsToNode(ctype, &(Nodes[i]), endpoint) ;
                   return TRUE ;
               }
               j++ ;
               parent = i ;
               i = Nodes[i].Child ;
           } else {
               /*** Attach chain of remaining n-j nodes to parent ***/

               if (*nodes + n-j > MAXNODES){
                   /* return since not enough memory for tree */  
                   return FALSE ;
               }
               
               /* Create nodes */
               for (t= *nodes ; t < *nodes + n-j ; t++) {
                   Nodes[t].Site = Walk[t- *nodes +j] ;
                   InitNodeCounts(&(Nodes[t])) ;
                   Nodes[t].Child = t+1 ;
                   Nodes[t].Sibling = DUMMY ;
                   Nodes[t].Parent = t-1 ;
               }

               /* Adjust top of chain */
               if (i == DUMMY)
                   Nodes[parent].Child = *nodes ;
               Nodes[*nodes].Parent = parent ;

               /* Adjust bottom of chain */
               Nodes[*nodes + n-j-1].Child = DUMMY ;
               if (ctype == 1){
                   Nodes[*nodes + n-j-1].Count++ ;
               } else if (ctype == 2){
                   Nodes[*nodes + n-j-1].Count2++ ;
               }
               AddCountsToNode(ctype, &(Nodes[*nodes + n-j-1]), endpoint) ;
               
               /* Insert top node into sibling list */
               if (i != DUMMY){
                   if(Nodes[i].Site < Walk[j]){
                       /* Nodes[i].Sibling == DUMMY */
                       Nodes[i].Sibling = *nodes ;
                   } else {
                       /* Nodes[i].Site > Walk[j] */
                       if (previ == i){
                           /* top node is first in the sibling list */
                           Nodes[parent].Child = *nodes ;
                           Nodes[*nodes].Sibling = i ;
                       } else {
                           Nodes[previ].Sibling = *nodes ;
                           Nodes[*nodes].Sibling = i ;
                       }
                   }
               }

               *nodes += n-j ;
               return TRUE ;
           }
       }

       return TRUE ;
                     
} /* end AddWalkToTree */    


int FilterSort_Walk(int *Walk, int N, int *Term, int nt, int a, int b, int c){

    /* This function is a low-pass filter. It filters and sorts the walk 
       given by sites Walk[0], ..., Walk[N-1], N >=1, using the termination set
       of sites Term[nt-1] < ... < Term[0] of length nt and
       an interval [a,b] of site numbers, with 0 <= a <= b < Term[nt-1].

       The filter is passed only by a walk that contains the nt
       sites from Term and at least one site s with a <= s <= b.
       Otherwise, the function returns -1.  

       The filter can be based only on the termination set
       by setting a = -1.

       Sites s >= c are discarded from the walk.
       The output is sorted as Walk[0] < Walk[1] < ... < Walk[n-1].

       The function returns the new length n, which may be 0.
    */

       int i, t, tmin;

       /* Sort the walk */
       if (N > 1)
           InsertionSort(Walk, N) ;

       /* Check whether the walk contains all sites of Term */
       tmin = N;
       t = 0;
       for (i=nt-1; i>=0; i--){
           /* Find Term[i] in Walk */
           while (t<N && Walk[t]!=Term[i])
               t++;
           if (t==N)
               return -1 ; /* not found */

           /* Register tmin such that Walk[tmin] = Term[nt-1] */
           if (i==nt-1)
               tmin = t ;  
       }

       /* Check whether the walk contains a site from [a,b] */
       if (a != -1){
           for (t=tmin-1; t>=0; t--){
               if (Walk[t] <= b && Walk[t] >= a)
                   break ;
               else if (Walk[t] <= b)
                   return -1 ;
           }
       }

       /* Discard sites s >= c */
       for (t=N-1; t>=0; t--){
           if (Walk[t] < c)
               return t+1 ; /* Walk[0] ... Walk[t] remain */
       }

       return 0 ; /* nothing remains */

} /* end FilterSort_Walk */


void InsertionSort(int *A, int n) {

    /* This function sorts the array A of length n by increasing value */

       int i, j, tmp;

       for (i = 1; i < n; i++) {
            /* A[0] <= A[1] <= ... <= A[i-1] are in increasing order */
            j = i;
            tmp = A[i]; /* will be inserted in appropriate place */
            while (j>0 && A[j-1]>tmp) {
                A[j] = A[j-1];
                j--;
            }
            A[j] = tmp;
       }

} /* end InsertionSort */


struct zzout Correct (struct treenode *Nodes, int N, int countop, int *S, int ns,
               int symmS, nint *nincludes,
               struct bin *Bins, int maxbin,
               struct vertex *V, int *lattice, int *rl,
               int a, int b, int c) {

    /* This function returns the correction term of the total count 
       for all walks of length at most N that pass through the set of sites S.

       Input: Nodes = the data structure representing the original tree.
              N =  maximum walk length represented by the lattice.

              countop = operation to be executed on the counts:
                        countop = 1 means the contribution is count1*count1
                                  2 means the contribution is count1*count2

              S = S[0..ns-1] = subset of included sites of the current
              correction term. 

              nincludes is the total number of include operations on S 
              carried out so far, during the whole algorithm. 
              It is used as a time stamp. 
              
              Bins[0..maxbin] is an array of vertex bins. 
              Bins[s] contains all active tree vertices with site number s.
              Bins[s] is stored as a linked list with header vertex Bins[s].First
              and with the link of a vertex v given by Nodes[v].Sibling.
              maxbin is an upper bound:  site(v) <= maxbin for all vertices v. 
              The count of an active vertex v in the tree is the number of walks 
              leading up to v and going through all of S.
              Here, count concerns walks of Type 1, and count2 those of Type 2.
              
              V is an array used to store the counts of parents of active vertices
              just before modification, so that these values can be restored. 

              Symmetry is used: the square of the counts is 
              multiplied by the symmetry number symmS of the set S.
              If symmS = 0, the part of S equivalent to the terminal site S[0]
              is not the lexicografically largest.
              If symmS > 0, the symmetry number is known, and need not be computed
              anymore. Otherwise, symmS = DUMMY.

              a, b, c are site numbers (0 <= a <= b < c) used to filter sites
              to be included in S. S must include at least one site s
              with a <= s <= b and no sites with b < s < c.
              The filter is not applied if a = -1. 

       Output: Z = correction term for walk count 
               RR = correction term for squared end-to-end distance
    */

    cint count, count2 ;
    struct gridpoint total, total2 ;

    ccint Z=0, RR=0;
    struct zzout ZZout, zztmp ;

    nint v, parent, i, t;
    int s, smax, symmSnew, maxbin_new ;

    ZZout.Z = 0;
    ZZout.RR = 0;
    if (symmS==0)
        return ZZout ;
        

    /* Find maximum site number smax using the bin data structure */
    smax = ROOT ;
    for (s=maxbin; s>=0; s--){
        if (Bins[s].Stamp == *nincludes){
            /* bin has been created in this round */
            smax = s ;
            break ;
        }
    }

    if (smax == ROOT)
        return ZZout ; /* no active vertices */

    if (a != -1 && smax < a && S[ns-1] > b) {
        /* no site from [a,b] can be included in S */
        return ZZout ;
    }

    /*** Phase 1: Update the sum for subsets S excluding smax
                  but possibly including smaller sites ***/

    /* Compute new symmetry number if excluding smax makes the number final */
    symmSnew = symmS ;
    if (symmS == DUMMY && ns >= 1 && (smax == 0 || !Equiv(smax-1, S[0])))
        symmSnew = symmset(N, S, ns, lattice, rl) ;

    /* Handle the non-root parents of active vertices v with site[v] = smax.
       By construction, all parents are different.
       Inactive parents become active and are moved into their bin. */
       
    v = Bins[smax].First ; /* first available vertex with site smax */
    t = 0 ; /* number of parents */
    while (v != DUMMY){

        parent = Nodes[v].Parent ;
        if (Nodes[parent].Site != ROOT){
            /* Register vertex and old count of parent for later undoing */
            V[t].v = v ;
            CopyNodeCountsToVertex(&(Nodes[parent]), &(V[t])) ;

            /* Check if parent is already active, i.e. its stamp equals nincludes.
               The time stamp of the node is stored in the child.  */
            if (Nodes[parent].Child == *nincludes){
                 /* Add the count of v to its parent */
                 AddNodeCounts(&(Nodes[parent]), &(Nodes[v])) ;
            } else {
                 /* Assign the count of v to its parent */
                 CopyNodeCounts(&(Nodes[parent]), &(Nodes[v])) ;

                 /* Insert parent at header in current bin data structure */
                 InsertInBin (Nodes, Bins, parent, *nincludes) ;
            }
            t++ ;
        }
        v = Nodes[v].Sibling ;
    }

    zztmp = Correct (Nodes, N, countop,S, ns, symmSnew, nincludes,
                     Bins, smax-1, &V[t], lattice, rl, a, b, c) ;

    ZZout.Z += zztmp.Z ;
    ZZout.RR += zztmp.RR ;

    /* Restore the previous state  of the counts */
    for (i=t-1; i>=0; i--){
        v = V[i].v ;
        parent = Nodes[v].Parent ;
        CopyNodeCountsFromVertex(&(Nodes[parent]), &(V[i])) ;
    }
     
    if (a != -1 && b < smax && smax < c) {
        /* smax cannot be included in S */
        return ZZout ;
    }


    /*** Phase 2: Update the sum for subsets S including smax
                  and possibly smaller sites ***/

    S[ns] = smax ;
    count = 0 ;
    count2 = 0 ;
    InitPoint(&total);
    InitPoint(&total2);

    /* Recompute new symmetry number if smax is equivalent to S[0] */
    if (symmS == DUMMY && symmSnew != DUMMY && Equiv(smax,S[0]))
        symmSnew = symmset(N, S, ns+1, lattice, rl) ;

    /* Create the set of non-root parents of active vertices v
       with site[v] = smax. By construction, all parents are different.
       The parents become the new set of active vertices. */
       
    (*nincludes)++ ; /* increase time stamp and empty the bins */
    v = Bins[smax].First ; /* first available vertex with site smax */
    maxbin_new = 0 ;
    t = 0; /* number of parents */
    while (v != DUMMY){
        /* Register the count */
        count += Nodes[v].Count ;
        count2 += Nodes[v].Count2 ;
        AddCountsToPoint(1, &(Nodes[v]), &total) ;
        AddCountsToPoint(2, &(Nodes[v]), &total2) ;
        
        parent = Nodes[v].Parent ;
        if (Nodes[parent].Site != ROOT){

           /* Register vertex and old count of parent */
            V[t].v = v ;
            CopyNodeCountsToVertex(&(Nodes[parent]), &(V[t])) ;

            /* Assign the count of v to its parent */
            CopyNodeCounts(&(Nodes[parent]), &(Nodes[v])) ;

            s = Nodes[parent].Site ;
            if (s > maxbin_new)
                maxbin_new = s;
        
            /* Insert parent at header in bin data structure */
            InsertInBin (Nodes, Bins, parent, *nincludes) ;
        }
        v = Nodes[v].Sibling ;
    }

    zztmp = Correct (Nodes, N, countop, S, ns+1, symmSnew, nincludes,
                     Bins, maxbin_new, &V[t], lattice, rl, a, b, c) ;
    ZZout.Z -= zztmp.Z ;
    ZZout.RR -= zztmp.RR ;

    /* Restore the previous state  of the counts */
    for (i=t-1; i>=0; i--){
        v = V[i].v ;
        parent = Nodes[v].Parent ;
        CopyNodeCountsFromVertex(&(Nodes[parent]), &(V[i])) ;
    }
    
    /* Update the sum for subsets S including smax but no smaller sites.
       count = total count for active vertices v with site[v] = smax */
    if (countop==1){
        /* Compute Z and RR while avoiding overflow */
        Z = count ;
        Z *= count ; /* Z = count^2 */
        RR = count ; 
        RR *= total.rr ;
        RR -= Inprod(total, total) ;
        RR *= 2 ; /* RR = 2 * (count*total.rr - total^2) */
    } else if (countop==2){
        ccint tmp1, tmp2 ;
        Z = count ;
        Z *= count2 ; /* Z = count*count2 */
        tmp1 = count ; 
        tmp1 *= total2.rr ;
        tmp2 = count2 ; 
        tmp2 *= total.rr ;
        RR = tmp1 + tmp2 - 2*Inprod(total, total2) ;
        /* RR = count*total2.rr + count2*total.rr - 2*Inprod(total, total2) */
    }

    if (symmSnew != DUMMY) {
        Z *= symmSnew ;
        RR *= symmSnew ;
    } else {
        int tmpsymm ;
        tmpsymm = symmset(N, S, ns+1, lattice, rl) ;
        Z *= tmpsymm ;
        RR *= tmpsymm ;
    }
    ZZout.Z  += Z ;
    ZZout.RR += RR ;

    return ZZout;
}


void InitBins(struct treenode *Nodes, nint nodes,
              struct bin *Bins, int *maxbin){

    /* This function initialises the bin data structure of the tree representing 
       the self-avoiding walks. On output, Bins[s] contains the active vertices
       (tree nodes) with site number s, organised as a linked list.
       The active vertices are the vertices with nonzero counts.

       On input, maxbin is the last bin to be initialised.
       On output, maxbin is the last non-empty bin.
                      
       On output, the active vertices and their bins have time stamp 0;
       the others have timestamp -1.
    */
 
       nint i ;
       int s, maxbin_new=0 ;
       
       /* Initialise bin data structure to dummy values */
       for (s=0; s <= *maxbin; s++){
           Bins[s].First = DUMMY ;
           Bins[s].Stamp =  -1 ;
       }

       /* Initialise tree nodes to dummy values, for fields not initialised
          by InitTree */
       for (i=0; i<nodes; i++){
           Nodes[i].Sibling = DUMMY ;
           Nodes[i].Child = -1 ;
       }

       /*** Insert all active vertices into the bin data structure ***/ 
       for (i=1; i<nodes; i++){
           if (Nodes[i].Count > 0 || Nodes[i].Count2 > 0){
               /* Insert vertex i at header */
               InsertInBin(Nodes, Bins, i, 0);
               s = Nodes[i].Site ;
               if (s > maxbin_new)
                   maxbin_new = s ;
           }
       }

       *maxbin = maxbin_new ;

} /* end InitBins*/
 

void InsertInBin (struct treenode *Nodes, struct bin *Bins, nint i, nint stamp){

    /* This function inserts vertex i at the header of the linked list in Bins[s]
       for s = site(i), and it leaves a time stamp on Nodes[i] and Bins[s]. */

       int s ;

       s = Nodes[i].Site ;
       if (Bins[s].Stamp != stamp){
           Bins[s].First = DUMMY ; /* start with an empty bin */
           Bins[s].Stamp = stamp ;
       }
       
       Nodes[i].Child = stamp ;
       Nodes[i].Sibling = Bins[s].First ;
       Bins[s].First = i ;

} /* end InsertInBin */


void print_verylong(FILE *fp,ccint Z){

    /* This function prints a very long integer Z in chunks of 9 decimal digits 
       to a file pointed to by fp */

    int i, j, Chunk[NCHUNKS] ;

    if (Z == 0){
        fprintf(fp,"0") ;
        return ;
    }

    if (Z < 0){
        fprintf(fp,"-") ;
        Z = -Z ;
    }

    /* Retrieve chunks of 9 digits from Z */
    for (i=0; i<NCHUNKS; i++){
          Chunk[i] = Z%GIGA ;
          Z /= GIGA ;
    }

    /* Find the most significant nonzero chunk and print it */
    for (i=NCHUNKS-1; i>=0; i--)
        if (Chunk[i] > 0){
            fprintf(fp,"%d",Chunk[i]);
            break ;
        }

    /* Print the remainder, padding the left of each chunk with zeros */
    for (j=i-1; j>=0; j--)
        fprintf(fp,"%09d",Chunk[j]);

} /* end print_verylong */
