#include "saw.h"
#include "sawdoubler.h"
#include "symmetry.h"

/* These are the 3D symmetry dependent functions for self-avoiding walks,
   valid for simple cubic, body-centred cubic and face-centred cubic lattices. */

int val(int i, int j, int k, int dir, int N, int *lattice) {

  /* This function returns the site number of lattice point
     (i, j, k) taking symmetry into account, where 
     -N <= i, j, k <= N and |i|+|j|+|k| <= N.

     dir is a direction, but it is not used here.
     On input, it can be set to 0.

     Let V be the set of lattice points (i, j, k) defined by 
     0 <= i <= j <= k <= N and i+j+k <= N.
     
     On input, all points in the set V must already have been set, 
     so that lattice[r] is the site number of point (i, j, k) 
     with r = (i+N)*L*L + (j+N)*L + k+N, where L=2N+1.
     These site numbers must be multiples of SYMM.
     
     The site number of a general point (i, j, k) equals the site number of its 
     equivalent in V plus a number in the range 0..SYMM-1 denoting the symmetry 
     operations needed to transform (i, j, k) into that equivalent.
  */
     
     int L ;
     
     L = 2*N + 1 ;
     
     /*** Reflections ***/

     if (i<0) return val(-i, j, k, dir, N, lattice) + 1;
     if (j<0) return val( i,-j, k, dir, N, lattice) + 2;
     if (k<0) return val( i, j,-k, dir, N, lattice) + 4;
   
     /*** Permutations ***/

     /* k = max(i,j,k) */
     if ((i<=j)&&(j<=k)) return lattice[((i+N)*L + j+N)*L + k+N] ;
     if ((j< i)&&(i<=k)) return lattice[((j+N)*L + i+N)*L + k+N] + REFL ;

     /* j = max(i,j,k) */
     if ((i<=k)&&(k<=j)) return lattice[((i+N)*L + k+N)*L + j+N] + 2*REFL ;
     if ((k< i)&&(i<=j)) return lattice[((k+N)*L + i+N)*L + j+N] + 3*REFL ;

     /* i = max(i,j,k) */
     if ((j<=k)&&(k<=i)) return lattice[((j+N)*L + k+N)*L + i+N] + 4*REFL ;
     if ((k< j)&&(j<=i)) return lattice[((k+N)*L + j+N)*L + i+N] + 5*REFL ;

     return DUMMY ; /* The function should not reach this point */

} /* end val */


int symmset(int N, int *S, int ns, int *lattice, int *rl) {

  /* This function returns either the symmetry number of set S of lattice points,
     or 0.  The symmetry number is the number of different sets of points that can be
     obtained from S' using permutations and reflections, where
         S' = {s in S : s equivalent to S[0]}.

     The set S has ns points, which are assumed to be ordered by decreasing
     site number. 

     The symmetry number is returned only if the set S' is the lexicographic
     largest among the different sets. Otherwise, 0 is returned.
  */

     int x[MAXN], y[MAXN], z[MAXN] ; /* 3D coordinates of S'*/
     int S2[MAXN] ; /* permuted/reflected set S' */
     int i, j, k, L, L2, s, T0, ri, xi, yi, zi, ns1, nequal ;

     if (ns <= 0)
         return 0 ;

     L = 2*N + 1 ;
     L2 = L*L ;

     /* Determine S' from S */
     T0 = (S[0]/SYMM)*SYMM ; /* S' = {s in S : s >= T0 } */

     for (i = 1; i < ns; i++) /* no need to check S[0] */
         if (S[i] < T0)
             break ; 
     ns1 = i ; /* S' = S[0..ns1-1] */

     /* Translate sites of S' to 3D coordinates */
     for (i = 0; i < ns1; i++){
         ri = rl[S[i]] ;
         x[i] = ri/L2 ;
         ri -= x[i]*L2 ;
         y[i] = ri/L ;
         z[i] = ri - y[i]*L ;
         x[i] -= N; y[i] -= N; z[i] -= N;
     }

     /* Determine all permuted and reflected sets that can be obtained from S' */
     nequal = 0 ; /* number of such sets equal to S' */
     for (k = 0; k < SYMM/REFL; k++){
         for (j = 0; j < REFL; j++){
             /* Perform symmetry operation k*REFL+j on set S' */
             for (i = 0; i < ns1; i++){

                 xi = x[i]; yi = y[i]; zi = z[i];

                 /* First execute reflection j and then permutation k */
                 reflect(&xi, &yi, &zi, j) ;
                 permute(&xi, &yi, &zi, k) ;
                 S2[i] = lattice[((xi+N)*L + yi+N)*L + zi+N] ;
             }

             /* Sort in increasing order */
             InsertionSort(S2, ns1) ;

             /* Lexicographically compare S2 and S' */
             for (i = 0; i < ns1; i++){
                 s = S2[ns1-1-i] ; /* to use S2 in decreasing order */
                 if (s>S[i]){
                     /* S' is not the lexicographic largest */
                     return 0 ;
                 } else if (s<S[i]){
                     /* decreasing S2 is lexicographically smaller than S' */
                     break ;
                 } 
             }

             if (i == ns1) /* decreasing S2 equals S' */
                 nequal++ ;
         }
     }

     /* Let G be the group of symmetry operations of the 3D cubic lattice.
        Let H be the subgroup of operations that leave S' invariant.
        Then SYMM = |G| and nequal = |H|. By Lagrange's theorem,
        nequal divides SYMM. */

     return SYMM/nequal ;

} /* end symmset */


int Canon(int N) {

    /* Computes number of lattice sites in the canonical numbering */
    
    int m ;

    m = 2*N + 1 ;
    return m*m*m ;

}  /* end Canon */
